varying vec3 light_dir;
varying vec3 eye_dir;

uniform vec3 diffuse_color;
uniform vec3 specular_color;
uniform float shininess;

uniform float bump_density;
uniform float bump_size;

void main(void)
{
    vec2 c = bump_density * gl_TexCoord[0].st;
    vec2 p = fract(c) - vec2(0.5);
    
    float d = dot(p, p);
    float f = 1.0 / sqrt(d + 1.0);

    if (d >= bump_size) {
	p = vec2(0.0);
	f = 1.0;
    }
    
    vec3 normal_delta = vec3(p.x, p.y, 1.0) * f;
    vec3 reflect_dir = reflect(normalize(light_dir), normal_delta);

    float Kd = max(dot(normal_delta, normalize(light_dir)), 0.0);

    float Ks = max(dot(normalize(eye_dir), reflect_dir), 0.0);
    Ks = pow(Ks, shininess);

    vec3 color = min((Kd * diffuse_color) + (Ks * specular_color),
		     vec3(1.0));

    gl_FragColor = vec4(color, 1.0);
}
