/*
 * (C) Copyright Ian D. Romanick 2007
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * AUTHORS, COPYRIGHT HOLDERS, AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
# include <GL/gl.h>

/* You many need to change this to "glext.h" depending on your build
 * environment.
 */
# include <GL/glext.h>

# define GET_PROC_ADDRESS(s) wglGetProcAddress((LPCSTR) s)
#else
# include <GL/glx.h>
# include <GL/glxext.h>

# define GET_PROC_ADDRESS(s) glXGetProcAddress((const GLubyte *) s)

# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <unistd.h>
#endif

#include <stdio.h>


PFNGLDELETEOBJECTARBPROC gl_DeleteObject_ARB;
PFNGLGETHANDLEARBPROC gl_GetHandle_ARB;
PFNGLDETACHOBJECTARBPROC gl_DetachObject_ARB;
PFNGLCREATESHADEROBJECTARBPROC gl_CreateShaderObject_ARB;
PFNGLSHADERSOURCEARBPROC gl_ShaderSource_ARB;
PFNGLCOMPILESHADERARBPROC gl_CompileShader_ARB;
PFNGLCREATEPROGRAMOBJECTARBPROC gl_CreateProgramObject_ARB;
PFNGLATTACHOBJECTARBPROC gl_AttachObject_ARB;
PFNGLLINKPROGRAMARBPROC gl_LinkProgram_ARB;
PFNGLUSEPROGRAMOBJECTARBPROC gl_UseProgramObject_ARB;
PFNGLVALIDATEPROGRAMARBPROC gl_ValidateProgram_ARB;
PFNGLUNIFORM1FARBPROC gl_Uniform1f_ARB;
PFNGLUNIFORM2FARBPROC gl_Uniform2f_ARB;
PFNGLUNIFORM3FARBPROC gl_Uniform3f_ARB;
PFNGLUNIFORM4FARBPROC gl_Uniform4f_ARB;
PFNGLUNIFORM1IARBPROC gl_Uniform1i_ARB;
PFNGLUNIFORM2IARBPROC gl_Uniform2i_ARB;
PFNGLUNIFORM3IARBPROC gl_Uniform3i_ARB;
PFNGLUNIFORM4IARBPROC gl_Uniform4i_ARB;
PFNGLUNIFORM1FVARBPROC gl_Uniform1fv_ARB;
PFNGLUNIFORM2FVARBPROC gl_Uniform2fv_ARB;
PFNGLUNIFORM3FVARBPROC gl_Uniform3fv_ARB;
PFNGLUNIFORM4FVARBPROC gl_Uniform4fv_ARB;
PFNGLUNIFORM1IVARBPROC gl_Uniform1iv_ARB;
PFNGLUNIFORM2IVARBPROC gl_Uniform2iv_ARB;
PFNGLUNIFORM3IVARBPROC gl_Uniform3iv_ARB;
PFNGLUNIFORM4IVARBPROC gl_Uniform4iv_ARB;
PFNGLUNIFORMMATRIX2FVARBPROC gl_UniformMatrix2fv_ARB;
PFNGLUNIFORMMATRIX3FVARBPROC gl_UniformMatrix3fv_ARB;
PFNGLUNIFORMMATRIX4FVARBPROC gl_UniformMatrix4fv_ARB;
PFNGLGETOBJECTPARAMETERFVARBPROC gl_GetObjectParameterfv_ARB;
PFNGLGETOBJECTPARAMETERIVARBPROC gl_GetObjectParameteriv_ARB;
PFNGLGETINFOLOGARBPROC gl_GetInfoLog_ARB;
PFNGLGETATTACHEDOBJECTSARBPROC gl_GetAttachedObjects_ARB;
PFNGLGETUNIFORMLOCATIONARBPROC gl_GetUniformLocation_ARB;
PFNGLGETACTIVEUNIFORMARBPROC gl_GetActiveUniform_ARB;
PFNGLGETUNIFORMFVARBPROC gl_GetUniformfv_ARB;
PFNGLGETUNIFORMIVARBPROC gl_GetUniformiv_ARB;
PFNGLGETSHADERSOURCEARBPROC gl_GetShaderSource_ARB;
PFNGLBINDATTRIBLOCATIONARBPROC gl_BindAttribLocation_ARB;
PFNGLGETACTIVEATTRIBARBPROC gl_GetActiveAttrib_ARB;
PFNGLGETATTRIBLOCATIONARBPROC gl_GetAttribLocation_ARB;
PFNGLVERTEXATTRIBPOINTERARBPROC gl_VertexAttribPointer_ARB;
PFNGLENABLEVERTEXATTRIBARRAYARBPROC gl_EnableVertexAttribArray_ARB;
PFNGLDISABLEVERTEXATTRIBARRAYARBPROC gl_DisableVertexAttribArray_ARB;

bool init_glsl_glue(void)
{
    gl_DeleteObject_ARB = (PFNGLDELETEOBJECTARBPROC)
	GET_PROC_ADDRESS("glDeleteObjectARB");
    if (gl_DeleteObject_ARB == NULL) {
	return false;
    }

    gl_GetHandle_ARB = (PFNGLGETHANDLEARBPROC)
	GET_PROC_ADDRESS("glGetHandleARB");
    if (gl_GetHandle_ARB == NULL) {
	return false;
    }

    gl_DetachObject_ARB = (PFNGLDETACHOBJECTARBPROC)
	GET_PROC_ADDRESS("glDetachObjectARB");
    if (gl_DetachObject_ARB == NULL) {
	return false;
    }

    gl_CreateShaderObject_ARB = (PFNGLCREATESHADEROBJECTARBPROC)
	GET_PROC_ADDRESS("glCreateShaderObjectARB");
    if (gl_CreateShaderObject_ARB == NULL) {
	return false;
    }

    gl_ShaderSource_ARB = (PFNGLSHADERSOURCEARBPROC)
	GET_PROC_ADDRESS("glShaderSourceARB");
    if (gl_ShaderSource_ARB == NULL) {
	return false;
    }

    gl_CompileShader_ARB = (PFNGLCOMPILESHADERARBPROC)
	GET_PROC_ADDRESS("glCompileShaderARB");
    if (gl_CompileShader_ARB == NULL) {
	return false;
    }

    gl_CreateProgramObject_ARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
	GET_PROC_ADDRESS("glCreateProgramObjectARB");
    if (gl_CreateProgramObject_ARB == NULL) {
	return false;
    }

    gl_AttachObject_ARB = (PFNGLATTACHOBJECTARBPROC)
	GET_PROC_ADDRESS("glAttachObjectARB");
    if (gl_AttachObject_ARB == NULL) {
	return false;
    }

    gl_LinkProgram_ARB = (PFNGLLINKPROGRAMARBPROC)
	GET_PROC_ADDRESS("glLinkProgramARB");
    if (gl_LinkProgram_ARB == NULL) {
	return false;
    }

    gl_UseProgramObject_ARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
	GET_PROC_ADDRESS("glUseProgramObjectARB");
    if (gl_UseProgramObject_ARB == NULL) {
	return false;
    }

    gl_ValidateProgram_ARB = (PFNGLVALIDATEPROGRAMARBPROC)
	GET_PROC_ADDRESS("glValidateProgramARB");
    if (gl_ValidateProgram_ARB == NULL) {
	return false;
    }

    gl_Uniform1f_ARB = (PFNGLUNIFORM1FARBPROC)
	GET_PROC_ADDRESS("glUniform1fARB");
    if (gl_Uniform1f_ARB == NULL) {
	return false;
    }

    gl_Uniform2f_ARB = (PFNGLUNIFORM2FARBPROC)
	GET_PROC_ADDRESS("glUniform2fARB");
    if (gl_Uniform2f_ARB == NULL) {
	return false;
    }

    gl_Uniform3f_ARB = (PFNGLUNIFORM3FARBPROC)
	GET_PROC_ADDRESS("glUniform3fARB");
    if (gl_Uniform3f_ARB == NULL) {
	return false;
    }

    gl_Uniform4f_ARB = (PFNGLUNIFORM4FARBPROC)
	GET_PROC_ADDRESS("glUniform4fARB");
    if (gl_Uniform4f_ARB == NULL) {
	return false;
    }

    gl_Uniform1i_ARB = (PFNGLUNIFORM1IARBPROC)
	GET_PROC_ADDRESS("glUniform1iARB");
    if (gl_Uniform1i_ARB == NULL) {
	return false;
    }

    gl_Uniform2i_ARB = (PFNGLUNIFORM2IARBPROC)
	GET_PROC_ADDRESS("glUniform2iARB");
    if (gl_Uniform2i_ARB == NULL) {
	return false;
    }

    gl_Uniform3i_ARB = (PFNGLUNIFORM3IARBPROC)
	GET_PROC_ADDRESS("glUniform3iARB");
    if (gl_Uniform3i_ARB == NULL) {
	return false;
    }

    gl_Uniform4i_ARB = (PFNGLUNIFORM4IARBPROC)
	GET_PROC_ADDRESS("glUniform4iARB");
    if (gl_Uniform4i_ARB == NULL) {
	return false;
    }

    gl_Uniform1fv_ARB = (PFNGLUNIFORM1FVARBPROC)
	GET_PROC_ADDRESS("glUniform1fvARB");
    if (gl_Uniform1fv_ARB == NULL) {
	return false;
    }

    gl_Uniform2fv_ARB = (PFNGLUNIFORM2FVARBPROC)
	GET_PROC_ADDRESS("glUniform2fvARB");
    if (gl_Uniform2fv_ARB == NULL) {
	return false;
    }

    gl_Uniform3fv_ARB = (PFNGLUNIFORM3FVARBPROC)
	GET_PROC_ADDRESS("glUniform3fvARB");
    if (gl_Uniform3fv_ARB == NULL) {
	return false;
    }

    gl_Uniform4fv_ARB = (PFNGLUNIFORM4FVARBPROC)
	GET_PROC_ADDRESS("glUniform4fvARB");
    if (gl_Uniform4fv_ARB == NULL) {
	return false;
    }

    gl_Uniform1iv_ARB = (PFNGLUNIFORM1IVARBPROC)
	GET_PROC_ADDRESS("glUniform1ivARB");
    if (gl_Uniform1iv_ARB == NULL) {
	return false;
    }

    gl_Uniform2iv_ARB = (PFNGLUNIFORM2IVARBPROC)
	GET_PROC_ADDRESS("glUniform2ivARB");
    if (gl_Uniform2iv_ARB == NULL) {
	return false;
    }

    gl_Uniform3iv_ARB = (PFNGLUNIFORM3IVARBPROC)
	GET_PROC_ADDRESS("glUniform3ivARB");
    if (gl_Uniform3iv_ARB == NULL) {
	return false;
    }

    gl_Uniform4iv_ARB = (PFNGLUNIFORM4IVARBPROC)
	GET_PROC_ADDRESS("glUniform4ivARB");
    if (gl_Uniform4iv_ARB == NULL) {
	return false;
    }

    gl_UniformMatrix2fv_ARB = (PFNGLUNIFORMMATRIX2FVARBPROC)
	GET_PROC_ADDRESS("glUniformMatrix2fvARB");
    if (gl_UniformMatrix2fv_ARB == NULL) {
	return false;
    }

    gl_UniformMatrix3fv_ARB = (PFNGLUNIFORMMATRIX3FVARBPROC)
	GET_PROC_ADDRESS("glUniformMatrix3fvARB");
    if (gl_UniformMatrix3fv_ARB == NULL) {
	return false;
    }

    gl_UniformMatrix4fv_ARB = (PFNGLUNIFORMMATRIX4FVARBPROC)
	GET_PROC_ADDRESS("glUniformMatrix4fvARB");
    if (gl_UniformMatrix4fv_ARB == NULL) {
	return false;
    }

    gl_GetObjectParameterfv_ARB = (PFNGLGETOBJECTPARAMETERFVARBPROC)
	GET_PROC_ADDRESS("glGetObjectParameterfvARB");
    if (gl_GetObjectParameterfv_ARB == NULL) {
	return false;
    }

    gl_GetObjectParameteriv_ARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
	GET_PROC_ADDRESS("glGetObjectParameterivARB");
    if (gl_GetObjectParameteriv_ARB == NULL) {
	return false;
    }

    gl_GetInfoLog_ARB = (PFNGLGETINFOLOGARBPROC)
	GET_PROC_ADDRESS("glGetInfoLogARB");
    if (gl_GetInfoLog_ARB == NULL) {
	return false;
    }

    gl_GetAttachedObjects_ARB = (PFNGLGETATTACHEDOBJECTSARBPROC)
	GET_PROC_ADDRESS("glGetAttachedObjectsARB");
    if (gl_GetAttachedObjects_ARB == NULL) {
	return false;
    }

    gl_GetUniformLocation_ARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
	GET_PROC_ADDRESS("glGetUniformLocationARB");
    if (gl_GetUniformLocation_ARB == NULL) {
	return false;
    }

    gl_GetActiveUniform_ARB = (PFNGLGETACTIVEUNIFORMARBPROC)
	GET_PROC_ADDRESS("glGetActiveUniformARB");
    if (gl_GetActiveUniform_ARB == NULL) {
	return false;
    }

    gl_GetUniformfv_ARB = (PFNGLGETUNIFORMFVARBPROC)
	GET_PROC_ADDRESS("glGetUniformfvARB");
    if (gl_GetUniformfv_ARB == NULL) {
	return false;
    }

    gl_GetUniformiv_ARB = (PFNGLGETUNIFORMIVARBPROC)
	GET_PROC_ADDRESS("glGetUniformivARB");
    if (gl_GetUniformiv_ARB == NULL) {
	return false;
    }

    gl_GetShaderSource_ARB = (PFNGLGETSHADERSOURCEARBPROC)
	GET_PROC_ADDRESS("glGetShaderSourceARB");
    if (gl_GetShaderSource_ARB == NULL) {
	return false;
    }

    gl_BindAttribLocation_ARB = (PFNGLBINDATTRIBLOCATIONARBPROC)
	GET_PROC_ADDRESS("glBindAttribLocationARB");
    if (gl_BindAttribLocation_ARB == NULL) {
	return false;
    }

    gl_GetActiveAttrib_ARB = (PFNGLGETACTIVEATTRIBARBPROC)
	GET_PROC_ADDRESS("glGetActiveAttribARB");
    if (gl_GetActiveAttrib_ARB == NULL) {
	return false;
    }

    gl_GetAttribLocation_ARB = (PFNGLGETATTRIBLOCATIONARBPROC)
	GET_PROC_ADDRESS("glGetAttribLocationARB");
    if (gl_GetAttribLocation_ARB == NULL) {
	return false;
    }

    gl_VertexAttribPointer_ARB = (PFNGLVERTEXATTRIBPOINTERARBPROC)
	GET_PROC_ADDRESS("glVertexAttribPointerARB");
    if (gl_VertexAttribPointer_ARB == NULL) {
	return false;
    }

    gl_EnableVertexAttribArray_ARB = (PFNGLENABLEVERTEXATTRIBARRAYARBPROC)
	GET_PROC_ADDRESS("glEnableVertexAttribArrayARB");
    if (gl_EnableVertexAttribArray_ARB == NULL) {
	return false;
    }

    gl_DisableVertexAttribArray_ARB = (PFNGLDISABLEVERTEXATTRIBARRAYARBPROC)
	GET_PROC_ADDRESS("glDisableVertexAttribArrayARB");
    if (gl_DisableVertexAttribArray_ARB == NULL) {
	return false;
    }

    return true;
}


/**
 * Load shader program code from the named source file.
 *
 * \param file_name  Name of the source file to be loaded.
 *
 * \return
 * A pointer to \c NUL terminated program source code on success or \c NULL
 * on failure.
 */
GLcharARB *glsl_load_shader_code(const char *file_name)
{
    GLcharARB *text = NULL;

#if defined(_WIN32)
    FILE *fp = fopen(file_name, "r");

    if (fp == NULL) {
	return NULL;
    }
    
    if (fseek(fp, 0, SEEK_END) == 0) {
	long len = ftell(fp);
	rewind(fp);

	text = new(GLcharARB[len + 1]);
	if (text != NULL) {
	    size_t bytes = fread(text, 1, len, fp);

	    if (bytes < (size_t) len) {
		delete text;
		text = NULL;
	    }
	    else {
		text[len] = '\0';
	    }
	}
    }

    fclose(fp);
    return text;
#else
    struct stat st;
    int fd = open(file_name, O_RDONLY);

    if (fd < 0) {
	return NULL;
    }

    if (fstat(fd, & st) == 0) {
	ssize_t total_read = 0;

	text = new(GLcharARB[st.st_size + 1]);
	if (text != NULL) {
	    do {
		ssize_t bytes = read(fd, text + total_read, 
				     st.st_size - total_read);
		if (bytes < 0) {
		    delete text;
		    text = NULL;
		    break;
		}

		if (bytes == 0) {
		    break;
		}

		total_read += bytes;
	    } while (total_read < st.st_size);

	    text[total_read] = '\0';
	}
    }

    close(fd);

    return text;
#endif
}
