/*
 * (C) Copyright Ian D. Romanick 2007
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * AUTHORS, COPYRIGHT HOLDERS, AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifdef _WIN32
# include <windows.h>
#endif

#include <GL/gl.h>
#include "vector.h"

//#define DUMP_VECTORS

#ifdef DUMP_VECTORS
# include <stdio.h>
#endif

/**
 * Generate an array of tangent vectors for a given set of polygon data.
 * 
 * \note
 * This implementation is not complete and is not very good.
 */
void generate_tangent_vectors(GLenum mode, 
			      const GLfloat *positions,
			      const GLfloat *normals,
			      GLfloat *tangents,
			      unsigned count)
{
    (void) mode;

    for (unsigned i = 0; i < count; i += 4) {
	const idr_vec4 p0(positions);
	const idr_vec4 p1(positions + 4);
	const idr_vec4 p2(positions + 8);
	const idr_vec4 p3(positions + 12);
	const idr_vec4 n0(normals);
	const idr_vec4 n1(normals + 3);
	const idr_vec4 n2(normals + 6);
	const idr_vec4 n3(normals + 9);

	idr_vec4 e = (p2 - p0).normalize();
	idr_vec4 t = e.cross(n0);

	tangents[0] = t[0];
	tangents[1] = t[1];
	tangents[2] = t[2];
#ifdef DUMP_VECTORS
	printf("{ %.2f %.2f %.2f } { %.2f %.2f %.2f }\n",
	       n0[0], n0[1], n0[2], t[0], t[1], t[2]);
#endif

	t = e.cross(n1);

	tangents[3] = t[0];
	tangents[4] = t[1];
	tangents[5] = t[2];
#ifdef DUMP_VECTORS
	printf("{ %.2f %.2f %.2f } { %.2f %.2f %.2f }\n",
	       n1[0], n1[1], n1[2], t[0], t[1], t[2]);
#endif

	t = e.cross(n2);

	tangents[6] = t[0];
	tangents[7] = t[1];
	tangents[8] = t[2];
#ifdef DUMP_VECTORS
	printf("{ %.2f %.2f %.2f } { %.2f %.2f %.2f }\n",
	       n2[0], n2[1], n2[2], t[0], t[1], t[2]);
#endif

	t = e.cross(n3);

	tangents[9] = t[0];
	tangents[10] = t[1];
	tangents[12] = t[2];
#ifdef DUMP_VECTORS
	printf("{ %.2f %.2f %.2f } { %.2f %.2f %.2f }\n",
	       n3[0], n3[1], n3[2], t[0], t[1], t[2]);
#endif

	positions += (4 * 4);
	normals += (3 * 4);
	tangents += (3 * 4);
    }
}
